/* 
  newInnerControl.c
  A cleaner version of inner_control.c
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "path.h"
#include "option.h"
#include "mesh.h"
#include "util.h"
#include "blackScholes.h"

/* Routine that computes the estimate of the controls */
/* 
  whichControl := choice of controls
  theMesh     := mesh data
  t            := current time at which we want estimates and such
  cntrlEst     := mesh estimate of control
  cntrlExct    := exact values of control from independent means
*/
void ControlEstimate(int whichControl, meshData *theMesh, int t, \
                     int fromPath, double *cntrlEst, double *cntrlExct)
{
  int path, imax;
  double sigma, divrate;

  /* Select control */
  switch(whichControl){
    case 1:    /* * * * Control 1 in paper * * * */
        /* control here is
           E[exp(-r \delta t) max(argmax(S1,...,S5)-K)|S_t =X_t(i)]
           See equation 24 in paper on page 18.
        */
        /* --- Compute exact or ind value of control --- */
        /* Compute maximum value of underlier at time t! */
        imax = dWhichMax(theMesh->option.pData.underliers, \
                  theMesh->option.pData.price[t][fromPath]) ;

        /* Using max underlier at t use Black-Scholes to price... */
        /* Extract volatility for max underlier */
        sigma = sqrt(theMesh->option.pData.covariance[imax][imax]);
        *cntrlExct = euroCall( \
                     theMesh->option.pData.price[t][fromPath][imax], \
                     theMesh->option.strikePrice, \
                     theMesh->option.riskFreeRate, \
                     sigma, 1.0,  \
                     theMesh->option.pData.dividendRate[imax]);

        /* --- Compute Estimate of Control --- */
      /* Loop over paths at current time */
      for(path=0; path <  theMesh->meshPaths; path++){
        /* Use the max underlier at time t to get value at t+1 */
        cntrlEst[path] = dMax(theMesh->option.pData.price[t+1][path][imax]\
                              - theMesh->option.strikePrice, 0.0);
        /* Deflate the price using the riskfree interest rate */
        cntrlEst[path] *= exp(-theMesh->option.riskFreeRate); /* dt=1 */

      } /* End loop over paths */
      break;

    case 2:    /* * * Control 2 in paper * * */
        /* control here is
           E[max(argmax(S1,...,S5))|S_t =X_t(i)]
           See equation 25 in paper on page 18.
        */
        /* --- Compute exact or ind value of control --- */
        /* Compute maximum value of underlier at time t! */
        imax = dWhichMax(theMesh->option.pData.underliers, \
                  theMesh->option.pData.price[t][fromPath]) ;
        /* compute value from given formula in paper */
        *cntrlExct = theMesh->option.pData.price[t][fromPath][imax];
        *cntrlExct *= exp(theMesh->option.riskFreeRate \
                      -  theMesh->option.pData.dividendRate[imax]);

        /* --- Compute Estimate of Control --- */
      /* Loop over paths at current time */
      for(path=0; path <  theMesh->meshPaths; path++){
        /* Use the max underlier at time t to get value at t+1 */
        cntrlEst[path] = theMesh->option.pData.price[t+1][path][imax];

      } /* End loop over paths */
     break;
  } /* End of the switch over choices of controls */
}  /* End Control Estimate */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Routine to compute E[Q(t+1,X)|X=x] based on paper.. ie mesh estimate*/
double Expectation(double *density, double *Quantity, int b){
  int j;
  double value;
  value = 0.0;

  for(j=0;j<b;j++){
    value += density[j] * Quantity[j];
  } 
  return(value);
} /* end Expectation */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Routine to compute sum of an array */
double Sum(double *array, int size){
  int j;
  double value;
   value =0.0;
  for(j=0;j<size; j++){
    value += array[j];
  }
  return value;
} /* End Sum */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Routine that returns the product of two arrays */
double * Product(double *a, double *b, int size){
  int i;
  static double *c;
  if(c == NULL){
/*
   printf("Allocating memory for c\n");
*/
   c = (double *)malloc(sizeof(double) * size);
  }
  for(i=0;i<size;i++)c[i] = a[i]*b[i];
  return(c);
}
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Routine to compute inner control variates */
/*
  whichControl = 1, 2 and 3 from paper.
*/
void ApplyInnerControl(int whichControl, meshData *theMesh){
  int i, t, b, whichMax; 
  double alpha, beta;
  double E_Q, sumW, E_nu, E_nu_Q, E_nu2;
  double *cntrlEst;   /* Array to contain control estimate */
  double cntrlExct;  /*  Known value of control */

  b = theMesh->meshPaths;

  /* Allocate control array */
  cntrlEst = (double *)malloc(sizeof(double)* \
                              theMesh->meshPaths);
  
  /* Loop over times starting from time T-1 in mesh */
  for(t=theMesh->option.pData.times-2;t>=0;t--){ 
  /* Note difference above b/w Scott's code. Zero included as before */
    /* Loop over all paths */
    /* doing backwards as in Scott's approach...*/
    for(i=theMesh->option.pData.paths-1;i>=0;i--){ 

      /* At time t at path i out of ALL paths look at time t+1 and 
         construct mesh estimates of controls there to get values at time
         t. These are to be "compared" to independent known values of the
         control at time t */

      /* Compute estimates and exact values of control */
      ControlEstimate(whichControl, theMesh, t,i, cntrlEst, &cntrlExct);

      /* ---Compute beta--- */
      /* Compute E[Q|...] */
      E_Q = Expectation(theMesh->weights[t][i], theMesh->value[t+1], b);
      /* Compute sum over weights */
      sumW = Sum(theMesh->weights[t][i],b);
      /* Compute E[nu|...] */
      E_nu =  Expectation(theMesh->weights[t][i], cntrlEst, b);
      /* Compute E[nu Q|...] */
      E_nu_Q = Expectation(theMesh->weights[t][i], \
                 Product(cntrlEst, theMesh->value[t+1],b), b);
      /* Compute E[nu nu |...] */
      E_nu2 = Expectation(theMesh->weights[t][i], \
                 Product(cntrlEst, cntrlEst,b), b);

      /* Evaluate beta from weighted least squares fit */
      beta = (E_nu_Q * sumW - E_nu * E_Q)/(E_nu2*sumW - E_nu*E_nu);
  
      /*--- Compute alpha--- */
      /* Maybe this is where I may have a problem */
      alpha = (E_Q - beta * E_nu) / sumW;

      /* Given the current value of the control at time t path i update
         with alpha and beta (based on mesh estimates of control at t+1)
         the estimates of Q at current time for current path */
      theMesh->newValueIfHeld[t][i] = alpha + beta * cntrlExct;


      /* Now recheck which is larger, valueIfHeld or payoff */
      if(theMesh->option.payoff[t][i] >  theMesh->newValueIfHeld[t][i]){
        theMesh->newValue[t][i] = theMesh->option.payoff[t][i];
      }else{
        theMesh->newValue[t][i] =  theMesh->newValueIfHeld[t][i];
/*
      printf("alpha = %lf, beta = %lf \n",alpha, beta);
      printf("mesh est c = %lf, exact c = %lf\n",E_nu, cntrlExct);
      printf("E_Q = %lf value = %lf\n",E_Q, theMesh->value[t][i]);
*/
      }

    } /* end loop over paths */
    
  } /* end loop over times */
  /* Free control arrays */
  free(cntrlEst);
} /* end ApplyInnerControl */
